#  ==========================================================================
#  normalize_toolkit_wip.rb
#  --------------------------------------------------------------------------
#
#  Description	:	Intended design of the script is to find manifold faces and
#					then normalize the faces depending on some parameters
#
#  Definition Of Use
#  -----------------
#
#  (1)	Test Manifold Algorithm
#
#		First select a group of edges, then select the menu 'Tools' ->
#		'Normalize Toolkit' -> 'Test manifold algorithm'
#
#  --------------------------------------------------------------------------
#  ==========================================================================
#  --------------------------------------------------------------------------
#
# Permission to use this software for any purpose and without fee is hereby granted
# Distribution of this software for commercial purpose is subject to:
#  - the expressed, written consent of the author
#  - the inclusion of the present copyright notice in all copies.
#
# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
#  --------------------------------------------------------------------------
#  ==========================================================================

module NormalizeToolkit


	class NormalizeFaces
		@@cursor_active = 0
		@@entityID = false
		@@tw = Sketchup.create_texture_writer
		@@t1 = 0
		@@t2 = 0
		@@t3 = 0
		@@t4 = 0
		@@t5 = 0
		@@t6 = 0
		@@difference_angle = 5.0
		
		@@vt_faces_in_surface_length		= "Waiting for input"
		@@vt_border_edge_length				= "..."
		@@vt_non_manifold_faces_length		= "..."
		@@vt_time_taken						= "..."
		def initialize
			#first lets check if we've got a selection of faces
			puts "test"
			selection = Sketchup.active_model.selection
			if selection.length <= 0
				self.alert_error("Error: No active selection. Please select the faces you want to normalize first and then run Normalize Surface.",true)
				Sketchup.send_action CMD_SELECT
				return false
			end #if
			
			entity_hash = self.get_entity_hash(selection)
			selection.clear
			@@faces_to_edges = entity_hash["faces_to_edges"]
			@@edges_to_faces = entity_hash["edges_to_faces"]
			@@faces_to_verts = entity_hash["faces_to_verts"]
			@@edges_to_verts = entity_hash["edges_to_verts"]
			if @@faces_to_edges.length <= 0
				self.alert_error("Error: No faces in selection. Please select the faces you want to normalize first and then run Normalize Surface.",true)
				Sketchup.send_action CMD_SELECT
				return false
			end #if
			
			# selection.clear
			# @@face_objects.each {|faceID, face|
				# selection.add face
			# }

			#now we load variables for the single face selection
			@@ph	= Sketchup.active_model.active_view.pick_helper
			Sketchup.vcb_label = "D.B.S."
			Sketchup.vcb_value = "5"
			self.load_cursors
			@@cursor = @@cursor_waiting
			Sketchup.active_model.active_view.refresh
		end #def
		
		def activate
			Sketchup.vcb_label = "Diff. Surface Angle:"
			Sketchup.vcb_value = @@difference_angle.to_s + " deg"
		end #def
		
		def resume(view)
			Sketchup.vcb_label = "Diff. Surface Angle:"
			Sketchup.vcb_value = @@difference_angle.to_s + " deg"	
		end #def
		
		def enableVCB?
			return true
		end #def
		
		
		def onUserText(text,view)
			angle = text.to_f
			if angle != 0
				@@difference_angle = angle
			end
			Sketchup.vcb_value = @@difference_angle.to_s + " deg"
		end #def
		
		def load_cursors
			@@cursor_specific	= UI::create_cursor((File.join(File.dirname(__FILE__),"normalize_toolkit","cursors","select_specific.png")),4,9)
			@@cursor_true		= UI::create_cursor((File.join(File.dirname(__FILE__),"normalize_toolkit","cursors","select_true.png")),4,9)
			@@cursor_waiting	= UI::create_cursor((File.join(File.dirname(__FILE__),"normalize_toolkit","cursors","waiting.png")),4,9)
			@@cursor_false		= UI::create_cursor((File.join(File.dirname(__FILE__),"normalize_toolkit","cursors","select_false.png")),4,9)
			@@cursor_active = @@cursor_specific
		end #def
		
		def alert_error(msg,critical=false,detail=false)
			puts ((detail != false) ? detail : msg)
			Sketchup.set_status_text msg, SB_PROMPT
			UI::messagebox(msg,MB_OK) if critical
		end #def
		
		def get_entity_hash(selection)
			faces_to_edges = Hash.new()							#these are associative arrays that connect types of
			faces_to_verts = Hash.new()							#entities by their physical connections
			edges_to_verts = Hash.new()
			edges_to_faces = Hash.new()
			@@face_objects = Hash.new()
			@@edge_objects = Hash.new()
			@@vert_objects = Hash.new()
			@@face_normals = Hash.new()
			for i in 0...selection.length
				if selection[i].kind_of?(Sketchup::Face)
					@@face_objects[selection[i].entityID] = selection[i]
					@@face_normals[selection[i].entityID] = selection[i].normal
					faces_to_edges[selection[i].entityID] = Array.new()
					edges = selection[i].edges
					for j in 0...edges.length
						if !edges_to_verts.has_key?(edges[j].entityID)
							edges_to_verts[edges[j].entityID] = Hash.new()
							verts = edges[j].vertices
							for k in 0...verts.length
								@@vert_objects[verts[k].entityID] = verts[k]
								edges_to_verts[edges[j].entityID][verts[k].entityID] = verts[k]
							end #for
						end #if
						@@edge_objects[edges[j].entityID] = edges[j]
						if !edges_to_faces.has_key?(edges[j].entityID)
							edges_to_faces[edges[j].entityID] = Array.new()
							faces = edges[j].faces
							for k in 0...faces.length
								edges_to_faces[edges[j].entityID] << faces[k].entityID
							end #for
						end #if
						faces_to_edges[selection[i].entityID] << edges[j].entityID
					end #for
					faces_to_verts[selection[i].entityID] = Array.new()
					verts = selection[i].vertices
					for j in 0...verts.length
						@@vert_objects[verts[j].entityID] = verts[j]
						faces_to_verts[selection[i].entityID] << verts[j].entityID
					end #for
				end #if
			end #for
			return Hash[
				"faces_to_edges" => faces_to_edges,
				"edges_to_faces" => edges_to_faces,
				"faces_to_verts" => faces_to_verts,
				"edges_to_verts" => edges_to_verts
			]
		end #def
		
		def onKeyUp(key, repeat, flags, view)
			case key
				when 27											#escape
					Sketchup.active_model.select_tool nil
					return false
			end #case
		end #def
		
		def onMouseMove(flags, x, y, view)
			selection = Sketchup.active_model.selection
			@@ph.do_pick x, y
			all_entities = @@ph.all_picked
			
			if(all_entities.length > 0)
				@@ph.do_pick x, y
				entity = @@ph.picked_face

				if entity != nil
					selection.clear if @@entityID != entity.entityID
					if @@faces_to_edges.has_key?(entity.entityID)
						#@@cursor_active = @@cursor_true
						view.tooltip = "Face in selection"
						if entity.entityID != @@entityID
							selection.add(entity)
							self.refine_surface_conditions(entity,true)
							@@entityID = entity.entityID
						end #if
					else
						@@vt_faces_in_surface_length		= "Waiting for input"
						@@vt_border_edge_length				= "..."
						@@vt_non_manifold_faces_length		= "..."
						@@vt_time_taken						= "..."
						@@cursor_active = @@cursor_false
						view.tooltip = "Face not in selection"
						Sketchup::set_status_text("Select a 2d manifold face within your previous selection",SB_PROMPT)
					end
				else
					@@vt_faces_in_surface_length		= "Waiting for input"
					@@vt_border_edge_length				= "..."
					@@vt_non_manifold_faces_length		= "..."
					@@vt_time_taken						= "..."
					selection.clear
					@@cursor_active = @@cursor_false
					view.tooltip = "No Face Found"
					@@entityID = false
					Sketchup::set_status_text("Select a 2d manifold face within your previous selection",SB_PROMPT)
				end
			else
				@@vt_faces_in_surface_length		= "Waiting for input"
				@@vt_border_edge_length				= "..."
				@@vt_non_manifold_faces_length		= "..."
				@@vt_time_taken						= "..."
				selection.clear
				@@cursor_active = @@cursor_specific
				view.tooltip = "No Face Found"
				@@entityID = false
				Sketchup::set_status_text("Select a 2d manifold face within your previous selection",SB_PROMPT)
			end #if
			
			view.refresh
		end #def
		
		def refine_surface_conditions(face,update_ui=false)
			d_angle_squared = @@difference_angle.degrees*@@difference_angle.degrees
			faces_to_check 	= Hash[face.entityID => face]		#current pass faces
			surface_group 	= Hash[face.entityID => face]		#hash to store manifold faces for each pass
			unchecked_faces = @@faces_to_edges.clone			#duplicate the selected faces and associated edges hash
			unchecked_edges = @@edges_to_faces.clone			#so we can remove checked faces and edges on each pass
			checked_faces 	= Hash.new()

			border_edges = Hash.new()							#hash to store & revise the border edges
			non_manifold_faces = Hash.new()						#hash to store non-manifold faces
			non_manifold_edges = Hash.new()						#hash to store non-manifold edges
			@@face_normals_reference = Hash.new()
			@@face_normals.each {|faceID, normVEC| 	@@face_normals_reference[faceID] = normVEC.clone;}
			
			@@vt_faces_in_surface_length		= "Calculating..."
			@@vt_border_edge_length				= "Calculating..."
			@@vt_non_manifold_faces_length		= "Calculating..."
			@@vt_time_taken						= "Calculating..."
			@@cursor = @@cursor_waiting
			Sketchup::set_status_text("Calculating...",SB_PROMPT)
			
			time = Time.now() if update_ui
			found_border_edges = false
			while !found_border_edges
				found_border_edges = true
				
				check_faces = faces_to_check.clone
				check_faces.each {|faceID, face|
					edges_arr = unchecked_faces[faceID]
					next if !edges_arr
					for i in 0...edges_arr.length
						edgeID = edges_arr[i]
						next if !unchecked_edges.has_key?(edgeID)
						faces_arr = unchecked_edges[edgeID].clone
						unchecked_edges.delete(edgeID)
						manifold_hash = self.find_manifold_face(faceID,edgeID,faces_arr,true)
						if !manifold_hash
							border_edges[edgeID] = @@edge_objects[edgeID]
						else
							non_manifold_edges[edgeID] = Hash.new() if !non_manifold_edges.has_key?(edgeID)
							manifold_face = manifold_hash["manifold_face"]
							manifold_faceID = manifold_face.entityID
							next if non_manifold_faces.has_key?(manifold_faceID)
							if manifold_hash["angle"]*manifold_hash["angle"] > d_angle_squared
								border_edges[edgeID] = @@edge_objects[edgeID]
								non_manifold_edges[edgeID]["border-manifold"] = true
								nf_array = faces_arr.clone
								nf_array.delete(manifold_faceID)
								nf_array.delete(faceID)
								non_manifold_faces.delete(faceID)
								nf_array.each{|nf_faceID| non_manifold_faces[nf_faceID] = true;}
							else
								if !manifold_hash["oriented"]
									border_edges[edgeID] = @@edge_objects[edgeID]
									non_manifold_faces[manifold_faceID] = true
									ne_array = @@faces_to_edges[manifold_faceID].clone
									#ne_array.delete(edgeID)
									ne_array.each {|ne_edgeID|
										if non_manifold_edges.has_key?(ne_edgeID)
											non_manifold_edges[ne_edgeID]["non-manifold"] = true
										else
											non_manifold_edges[ne_edgeID] = Hash["non-manifold" => true]
										end
									}
								else
									surface_group[manifold_faceID] 			= manifold_face
									faces_to_check[manifold_faceID]			= manifold_face	if !checked_faces.has_key?(manifold_face.entityID)
									non_manifold_edges[edgeID]["manifold"] 	= true
									nf_array = faces_arr.clone
									nf_array.delete(manifold_faceID)
									nf_array.delete(faceID)
									#nf_array.each {|nf_faceID| non_manifold_faces[nf_faceID] = true;}
								end #if
							end #if
						end
					end #for
					unchecked_faces.delete(faceID)
					faces_to_check.delete(faceID)
					checked_faces[faceID]	= true
					found_border_edges 		= false
				}
			end #while
			
			if surface_group.length > 1
				@@cursor_active = @@cursor_true
				Sketchup::set_status_text("Found Surface. Move to a new face to re-calculate surface",SB_PROMPT)
				non_manifold_faces_selection = Array.new()
				non_manifold_edges_selection = Array.new()
				non_manifold_faces.each_key {|faceID|	surface_group.delete(faceID); non_manifold_faces_selection << @@face_objects[faceID];}
				non_manifold_edges.each {|edgeID, edgeContext|
					manifold 		= edgeContext.has_key?("manifold")
					non_manifold 	= edgeContext.has_key?("non-manifold")
					border_manifold = edgeContext.has_key?("border-manifold")
					if (manifold && !border_manifold) || (non_manifold && !border_manifold)
						border_edges.delete(edgeID)
						non_manifold_edges_selection << @@edge_objects[edgeID]
					end #if		
				}
				if update_ui
					@@vt_faces_in_surface_length		= surface_group.length.to_s + " face(s)"
					@@vt_border_edge_length				= border_edges.length.to_s + " edge(s)"
					@@vt_non_manifold_faces_length		= non_manifold_faces.length.to_s + " face(s)"
					total_time = (((Time.now() - time)*1000).round).quo(1000)
					@@vt_time_taken						= total_time.to_s + " seconds"
					#puts "t1: "+@@t1.to_s+"\nt2: "+@@t2.to_s+"\nt3: "+@@t3.to_s+"\nt4: "+@@t4.to_s+"\nt5: "+@@t5.to_s+"\nt6: "+@@t6.to_s
					selection = Sketchup.active_model.selection
					selection.clear
					selection.add(border_edges.values)
					selection.add(non_manifold_faces_selection)
				end #if
				return Hash[
					"surface_group"	=> surface_group,
					"border_edges"	=> border_edges,
					"non_manifold_faces"	=> non_manifold_faces_selection,
					"non_manifold_edges"	=> non_manifold_edges_selection
				]
			else
				@@cursor_active = @@cursor_false
				Sketchup::set_status_text("Error (Only 1 Face Found). Move to a new face to re-calculate surface",SB_PROMPT)
				@@vt_faces_in_surface_length	= "1 face (Surface Aleady Normalized)"
				@@vt_border_edge_length			= border_edges.length.to_s + " edge(s)"
				@@vt_non_manifold_faces_length	= non_manifold_faces.length.to_s + " face(s)"
				total_time						= (((Time.now() - time)*1000).round).quo(1000)
				@@vt_time_taken					= total_time.to_s + " seconds"
				return false
			end	
		end #def
		
		def find_manifold_face(origin_face_id,fold_edge_id,possible_face_ids,check_orientation=false,check_angle=false)		
			origin_face		= @@face_objects[origin_face_id]
			origin_normal	= @@face_normals_reference[origin_face_id]
			origin_edges	= @@faces_to_edges[origin_face_id]
			manifold_face	= false
			
			time = Time.now()
			
			#	PT1: ORIGIN FACE->FOLD EDGE ORIENTATION
			#	we need to find on which side of the edge the face is, and then
			#	determine the direction of rotation. This is more difficult than
			#	it appears. The most error-free way to do it is to create a line
			#	that represents the planar normal for the edge (ie perpendicular
			#	to the fold edge and on the plane of the origin face) and find the
			#	line segments that intersect with the face using an Intersect Segment
			#	with Polygon algorithm. [see: http://softsurfer.com/Archive/algorithm_0111/algorithm_0111.htm]
			#	A COUPLE OF NOTES FIRST:
			# 1:We only need a 2d coordinate system for this so we can use the UVHelper
			#	to generate this and hopefully make the code more efficient
			# 2:We only need to find the first intersected line segment that shares
			#	one of its points with the fold edge
			
			#Mod 2: edge reversed_in? mod
			#Sketchup's reversed_in? method seems to determine the side of the edge that
			#the face is on. So if edge.reversed_in? face == true, then the face is on the
			#right hand side of the edge, if its false, then the face is on the lhs. This is
			#independent of outer loops or inner loops & respective of the face normal.
			
			fold_verts 		= @@edges_to_verts[fold_edge_id]					#construct points on the fold edge that will help with
			fold_arr		= @@edge_objects[fold_edge_id]						#orientation: pt1: start, pt2: middle, pt3: end
			fold_pt1		= fold_arr.start.position
			fold_pt3		= fold_arr.end.position
			fold_pt2		= fold_pt1 + Geom::Vector3d.new((fold_pt3.x-fold_pt1.x).quo(2),(fold_pt3.y-fold_pt1.y).quo(2),(fold_pt3.z-fold_pt1.z).quo(2))
			fold_vec		= fold_pt1.vector_to fold_pt3
			fold_reversed	= (@@edge_objects[fold_edge_id].reversed_in? origin_face) ? -1 : 1
			fold_reversed	= fold_reversed * -1 if !origin_normal.samedirection? origin_face.normal
			trans 			= Geom::Transformation.rotation fold_pt2, origin_normal, (90.degrees * fold_reversed)
			o_vec1 			= fold_vec.transform trans

			o_pt1 = fold_pt2 + o_vec1
			o_vec2 	= origin_normal										#make the origin face's normal an orientation vector
																		#we need two vectors at right angles to get around the
																		#angle_between not greater than 180 degrees problem.
			smallest_angle 	= 360.degrees
			@@t4 += (Time.now() - time)
			count = 0
			possible_face_ids.each {|faceID|
				next if !@@faces_to_verts.has_key?(faceID)
				next if faceID == origin_face_id
				#puts "edgeID: "+fold_edge_id.to_s+" faceID: "+origin_face_id.to_s if count == 0
				count +=1
				
				p_fold_reversed = (@@edge_objects[fold_edge_id].reversed_in? @@face_objects[faceID]) ? -1 : 1
				fold_reversed	= fold_reversed * -1 if !@@face_normals_reference[faceID].samedirection? @@face_objects[faceID].normal
				trans 			= Geom::Transformation.rotation fold_pt2, @@face_normals_reference[faceID], (90.degrees * p_fold_reversed)
				p_vec1			= fold_vec.clone
				p_vec1 			= p_vec1.transform trans
				
				o_ang1 = o_vec1.angle_between p_vec1
				o_ang2 = o_vec2.angle_between p_vec1
				
				manifold_angle = (o_ang2 <= 90.degrees) ? o_ang1 : 180.degrees + (180.degrees - o_ang1)
				#puts "pfold: "+p_fold_reversed.to_s+" fold: "+fold_reversed.to_s+" o_ang1: "+o_ang1.radians.to_s+" o_ang2: "+o_ang2.radians.to_s+" manifold angle: "+manifold_angle.radians.to_s

				if manifold_angle < smallest_angle
					#puts manifold_angle.radians
					smallest_angle = manifold_angle
					manifold_face = @@face_objects[faceID]
				end
			}
			@@t5 += (Time.now() - time)
			if !manifold_face
				return false
			end
			
			direction = self.get_rotation_direction(fold_vec,fold_pt2,origin_normal,o_pt1)					#work out whether the vector is pointing in the right direction
			trans 	= Geom::Transformation.rotation fold_pt2, fold_vec, (smallest_angle * direction)		#rotate the origin normal so its aligned with
			m_normal 	= o_vec2.transform trans															#the manifold's normal. This should give a vector
			m_ang1		= m_normal.angle_between @@face_normals_reference[manifold_face.entityID]			#that angles 180 degrees from the manifold's normal
			oriented = (m_ang1 > 90.degrees) ? true : false													#if its less than 90 degrees, reverse the face.
			if !oriented
				#puts "face reversed"
				@@face_normals_reference[manifold_face.entityID].reverse!
			end #if
			return Hash[
				"manifold_face" => manifold_face,
				"oriented"		=> oriented,
				"angle"			=> (180.degrees - smallest_angle)
			]
		end #def

		
		def get_rotation_direction(axis_vec,axis_point,rotation_vector,rotation_point)
			trans = Geom::Transformation.rotation axis_point, axis_vec, 90.degrees
			point = rotation_point.transform trans
			vec = axis_point.vector_to point
			
			angle = rotation_vector.angle_between vec
			direction = (angle > 90.degrees) ? -1 : 1
			return direction
		end #def
		
		def deactivate(view)
			puts "tool deactivated"
			view.invalidate
			Sketchup.active_model.selection.clear
			@@entityID = false
		end #def
		
		def suspend(view)
			#puts "tool suspended"
			view.invalidate
		end #def

		def onSetCursor
			UI.set_cursor(@@cursor_active)
		end #def

		
		def draw(view)
			v_height = view.vpheight
			status = view.draw2d GL_LINES, [Geom::Point3d.new(13,(v_height-145),0),Geom::Point3d.new(13,(v_height-145),0)]
			pt = Geom::Point3d.new 13,(v_height-145),0
			status = view.draw_text pt, "Statistics:"
			pt = Geom::Point3d.new 13,(v_height-130),0
			status = view.draw_text pt, "Faces Found: "+@@faces_to_edges.length.to_s + " faces"
			pt = Geom::Point3d.new 13,(v_height-115),0
			status = view.draw_text pt, "Faces In Surface: "+@@vt_faces_in_surface_length
			pt = Geom::Point3d.new 13,(v_height-100),0
			status = view.draw_text pt, "Border Edges Found: "+@@vt_border_edge_length
			pt = Geom::Point3d.new 13,(v_height-85),0
			status = view.draw_text pt, "Non-Manifold Faces Found: "+@@vt_non_manifold_faces_length
			pt = Geom::Point3d.new 13,(v_height-70),0
			status = view.draw_text pt, "Time Taken: "+@@vt_time_taken
			pt = Geom::Point3d.new 13,(v_height-55),0
			status = view.draw_text pt, "Legend:"
			pt = Geom::Point3d.new 13,(v_height-40),0
			status = view.draw_text pt, "Selected Faces: 2d manifold errors"
			pt = Geom::Point3d.new 13,(v_height-25),0
			status = view.draw_text pt, "Selected Edges: Border edge of surface"
		end #def
		
	end #class

end #module

#############################################
# this part adds the plugin option to the plugin menu
if not file_loaded?(File.basename(__FILE__))
  
	submenu = UI.menu("Tools").add_submenu("Normalize Toolkit")
	cmd = UI::Command.new("Test manifold algorithm") {
		@@tool = NormalizeToolkit::NormalizeFaces.new
		tool = Sketchup.active_model.select_tool @@tool
	}
	submenu.add_item(cmd)
	# submenu = UI.menu("Tools").add_submenu("Normalize Toolkit")
	# cmd = UI::Command.new("Normalize Single Surface") {
		# @@tool = NormalizeToolkit::NormalizeFaces.new
		# tool = Sketchup.active_model.select_tool @@tool
	# }
	# submenu.add_item(cmd)	
	# cmd = UI::Command.new("Normalize Multiple Surfaces") {
		# @@tool = NormalizeToolkit::NormalizeFaces.new
		# tool = Sketchup.active_model.select_tool @@tool
	# }
	# submenu.add_item(cmd)
end #if
file_loaded(File.basename(__FILE__))